package HyperEps.publish;

import java.security.InvalidAlgorithmParameterException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Random;
import java.util.Stack;

/**
 * Epsilon Subset Subset Selection 
 * expected runtime O((|A| log |B| + |B|) log (|A| |B|))
 * 
 * @author Patrick Klitzke <philologos14@gmail.com>
 * 
 * Reference as:
 * K. Bringmann, T. Friedrich, and P. Klitzke                                 
 *  Two-dimensional subset selection for hypervolume and  epsilon-indicator.                                                        
 * 16th Annual Conference on Genetic and Evolutionary  Computation Conference (GECCO~'14)}. ACM Press, 2014.                     
 * See http://docs.theinf.uni-jena.de/paper/2014GECCO.pdf
 */
public class EPSSSP {

  // private member of the class
  Point[] A; // the first point set
  Point[] B; // the second point set
  Random rand; // random variable

  /**
   * Take elements of B to approximate A
   * and deltes non dominateing points
   * @param P1
   * @param P2
   */
  public EPSSSP(Point[] A, Point[] B) {
    this.A = EPSSSP.makeNonDominant(A);
    this.B = EPSSSP.makeNonDominant(B);
    rand = new Random();
  }

  /**
   * Function that takes a set of Points and makes them non dominant. 
   * If b dominates a (i.e. a.x<b.x and a.y<b.y), then a is chosen 
   *
   * Runtime: O(n log n) because of sorting
   * @param front
   * @return
   */
  public static Point[] makeNonDominant(Point[] front){
    // sort the elements by increasing x-coordinate
    Arrays.sort(front, new Comparator<Point>(){
      @Override
      public int compare(Point o1, Point o2) {
        if (o1.x < o2.x) {
          return -1;
        } else  if (o1.x == o2.x){
          if(o1.y < o2.y) {
            return -1;
          } else if(o1.y == o2.y) {
            return 0;
          } else {
            return 1;
          }
        } else {
          return 1;
        }
      }
    });

    // go through from right and delete not dominating points
    Stack<Integer> st = new Stack<Integer>();
    st.push(front.length-1);

    for(int n=front.length-2; n>=0; n--){
      while(st.size() > 0 && front[st.peek()].y >= front[n].y) {
        st.pop();
      } 
      st.push(n);
    }

    // go through the stack and set it to the elements
    Point p[] = new Point[st.size()];
    for (int i=0; i < st.size(); i++) {
      p[st.size()-1-i] = front[st.get(i)];
    }

    return p;
  }

  /**
   * Indicator function I(a,b) for two points a and b. If I(a,b)=epsilon, that means that one has to move b by epsilon such that it dominates a.  
   * The definition is max(b.x-a.x, b.y - a.y) 
   * @param p1
   * @param p2
   * @return
   */
  public static double I(Point p1, Point p2) {
    return Math.max(
        p2.x - p1.x,
        p2.y - p1.y
        );
  }
  
  /**
   * This function performs the epsilon test. 
   * @return
   */
  public boolean epsilonTest(double epsilon, int k, boolean withExtrema) {
    boolean res = true;
    try{
      this.epsilonPoints(epsilon, k, withExtrema);
    }catch (InvalidAlgorithmParameterException ex)  {
      res = false;
    }

    return res;
  }

  public Point[] epsilonPoints(double epsilon, int k, boolean withExtrema) throws InvalidAlgorithmParameterException {
	  if (withExtrema) {
		  return epsilonPointsWithExtrema(epsilon, k);
	  } else {
		  return epsilonPointsWithoutExtrema(epsilon, k);
	  }
  }

  // returns the new Pointset with extrema
  public Point[] epsilonPointsWithExtrema(double epsilon, int k) throws InvalidAlgorithmParameterException {
    int posA = 0;
    int posB = 0;

    ArrayList<Point> newA = new ArrayList<>();

    for (int n=0; n<A.length; n++) {
      if (!(I(A[n], B[0]) <= epsilon || I(A[n], B[B.length-1]) <=epsilon) ) {
        newA.add(A[n]);
      }
    }

    k-=2;

    posB = 1;
    int bEnd = B.length-1;

    ArrayList<Point> p = new ArrayList<>();
    p.add(B[0]);

    posB = 1;

    while(posB < bEnd && posA < newA.size()) {
      k--;
      while (posB +1 < bEnd && B[posB+1].x - newA.get(posA).x <= epsilon  ) {
        posB++;
      }

      // delete all A points dominates by the point
      int help = posA;
      while (help < newA.size() && I(newA.get(help), B[posB]) <= epsilon) {
        help++;
      }

      p.add(B[posB]);

      if (help == posA) {
        // no approximation :(
        throw new InvalidAlgorithmParameterException();
      }
      posA = help;
    }

    p.add(B[B.length -1]);

    // return the sorted Points
    if (posA == newA.size()) {
      if(k >= 0) {
        Point[] points = new Point[p.size()];
        for(int n=0; n<p.size();n++) {
          points[n] = p.get(n);
        }
        return points;
      }
    }
    throw new InvalidAlgorithmParameterException();
  }

  /**
   * Return the points in EPSSSP without extrema 
   */
   public Point[] epsilonPointsWithoutExtrema(double epsilon, int k) throws InvalidAlgorithmParameterException {
    int posA = 0;
    int posB = 0;

    ArrayList<Point> p = new ArrayList<Point>();

    while(posB < B.length && posA < A.length) {
      k--;
      while (posB +1 < B.length && B[posB+1].x - A[posA].x <= epsilon  ) {
        posB++;
      }

      // delete all A points dominates by the point
      int help = posA;
      while (help < A.length && I(A[help], B[posB]) <= epsilon) {
        help++;
      }

      p.add(B[posB]);

      if (help == posA) {
        // no approximation :(
        throw new InvalidAlgorithmParameterException();
      }
      posA = help;
    }

    // return the sorted Points
    if (posA == A.length) {
      if(k >= 0) {
        Point[] points = new Point[p.size()];
        for(int n=0; n<p.size();n++) {
          points[n] = p.get(n);
        }
        return points;
      }
    }

    throw new InvalidAlgorithmParameterException();
  }


  /**
   * Returns the set data structure for the difference for each point b in B
   * runtime O(|A| + |B|)
   */
  private List<Set> createSet() {

    List<Set> v = new ArrayList<Set>();

    int a = -1;
    
    // create the set for every B
    for(Point pointB : B ) {
      double m = 1;
      double n = pointB.y - pointB.x; 

      // are these points still on the left?
      while(a+1 < A.length && (m*A[a+1].x + n) <= A[a+1].y) {
        a++;
      }

      // add the part left to b
      v.add(new Set(pointB, A, 0, a, true));
      
      // add the part right to a
      v.add(new Set(pointB, A, a+1, A.length-1, false));
    }

    return v;
  }

  /**
   * returns a random long in the range 0,1,..,(max-1)
   */
  public long randLong(long max) {
    long range = max;
    long number = (long)(rand.nextDouble()*range);
    return number;
  }

  /**
   * Fast solution for calculating the minimum epsilon for epsilon subset selection
   *  expected runtime O((|A| log |B| + |B|) log (|A| |B|))
   *  @param withExtrema: indicates whether we will select the extrema in the epsilon test
   */
  public double fastSolution(int k, boolean withExtrema) {
    List<Set> v = createSet();

    double best = Double.MAX_VALUE;

    int it =0;
    while(true) {
      it++;

      // get the size of all the sets
      long totalSize = 0;
      for (Set s : v) {
        totalSize += s.getSize();
      }

      // we are finished if we looked at all the sets
      if (totalSize == 0) {
        return best;
      }

      // choose a random valid distance and select it
      long num = randLong(totalSize);
      double val = 0;
      for(Set s : v) {
        if (s.getSize() <= num) {
          num -= s.getSize();
        } else {
          val = s.getElement((int)num);
          break;
        }
      }

      
      if (epsilonTest(val, k, withExtrema)) { // e <= e*

        if (val < best) {
          best = val;
        } 

        for (int n=0; n<v.size(); n++) {
          v.get(n).deleteBigger(best);
        }

        
      } else { // e >= e*
        for (int n=0; n<v.size(); n++) {
          v.get(n).deleteSmaller(val);
        }
      }

      // delete empty Sets
      ArrayList<Set> newList = new ArrayList<>();
      for(Set s : v) {
        if (s.getSize() > 0) {
          newList.add(s);
        }
      }
      v = newList;
    }
  }

  /**
   * Slow solution for calculating the minimal value for the epsilon indicator
   * Runtime: O(n^2)
   * @param k : how many points to select
   * @param withExtrema : using the extrema in the epsilon tests
   */
  public double slowSolution(int k, boolean withExtrema) {
    ArrayList<Double> possibleValues = new ArrayList<Double>();

    // calculate all the possible values 
    // Runtime : O(n^2)
    for (Point a : A) {
      for( Point b : B) {
        possibleValues.add(I(a,b));
      }
    }

    // sort the possible values so we can do binary search
    // Runtime O(n log n)
    Collections.sort(possibleValues);

    // binary search the values
    int begin = 0;
    int end = possibleValues.size() - 1;

    while(begin < end) {
      int mid = (begin + end) / 2;
      if (epsilonTest(possibleValues.get(mid), k, withExtrema)) {
        end = mid;
      } else {
        begin = mid + 1;
      }
    }
    return possibleValues.get(begin);
  }	
};
