package HyperEps.publish;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;

/**
 * Code for Hypervolume subset selection 
 * @author Patrick Klitzke <philologos14@gmail.com>
 * 
 * Reference as:
 * K. Bringmann, T. Friedrich, and P. Klitzke                                 
 *  Two-dimensional subset selection for hypervolume and  epsilon-indicator.                                                        
 * 16th Annual Conference on Genetic and Evolutionary  Computation Conference (GECCO~'14)}. ACM Press, 2014.                     
 * See http://docs.theinf.uni-jena.de/paper/2014GECCO.pdf
 */
public class HYPSSP {
	
	/**
	 * Calculation of the x intersection of two Functions f1 and f2. 
	 * @param f1 first function
	 * @param f2 second function
	 * @return
	 */
	 public static double IntersectX(Function f1, Function f2) {
		  return (f2.n - f1.n) / (f1.m - f2.m);
	  }
	 
	 // epsilon for doing floating point arithmetic
	 private static double EPSILON = 1e-10;
	 
	 /**
	  * Returns the position which funciton for each point is maximum. 
	  * i.e. max(F(i)[X[j]])= F[pos[j]](X[j])  
	  * The funcitons are ordered with decreasing slope
	  * @return
	  */
	 private static int[] CalcMaxPoint(Function[] f, int N, double[] X) {
	      int[] best = new int[N];
	      Function[] s = new Function[N];
	      int[] pos = new int[N];

	      int len = 0;
	      int left = 0;
	      int right = -1;

	      for (int i = N - 2; i >= 0; i--) {
	        while (len > 1 ) {
	          double d1 = IntersectX(f[i+1], s[right]);
	          double d2 = IntersectX(s[right], s[right - 1]);

	          if (d1 > d2 || Math.abs(d1-d2) < EPSILON) {
	            len--;
	            right--;
	          } else {
	            break;
	          }
	        }

	        s[right+1] = f[i+1];
	        pos[right+1] = i+1;
	        right++;
	        len++;

	        while (len > 1) {
	          double d1 = s[left].evaluate(X[i]);
	          double d2 = s[left + 1].evaluate(X[i]);

	          if (d1 < d2 || Math.abs(d1-d2) < EPSILON) {
	            len--;
	            left++;
	          } else {
	            break;
	          }
	        }

	        best[i] = pos[left];      
	       }

	      return best;
	    }
	 
	 /**
	  * Calculate the maximum values for each position X among the functions f.
	  */
	  private static double[] CalcMaxFast(Function[] f, int N, double[] X) {
	      double[] best = new double[N];
	      Function[] s = new Function[N];

	      int len = 0;
	      int left = 0;
	      int right = -1;

	      for (int i = N - 2; i >= 0; i--) {
	        while (len > 1 ) {
	          double d1 = IntersectX(f[i+1], s[right]);
	          double d2 = IntersectX(s[right], s[right - 1]);

	          if (d1 > d2 || Math.abs(d1-d2) < EPSILON) {
	            len--;
	            right--;
	          } else {
	            break;
	          }
	        }

	        s[right+1] = f[i+1];
	        right++;
	        len++;

	        while (len > 1) {
	          double d1 = s[left].evaluate(X[i]);
	          double d2 = s[left + 1].evaluate(X[i]);

	          if (d1 < d2 || Math.abs(d1-d2) < EPSILON) {
	            len--;
	            left++;
	          } else {
	            break;
	          }
	        }

	        best[i] = s[left].evaluate(X[i]);      
	      }

	      return best;
	    }
	  
	 
	   
	   /**
	    * Fast calculation O(n*k) for the hypervolume selection problem
	    * returns the selected points
	    */
	   private static Point[]  FasterPoint(int N, int K, double X[], double Y[]) {

	      // now it is time for the real calculation
	      Function[] fun = new Function[N];

	      for(int n=0; n<N; n++) {
	        fun[n] = new Function(-Y[n], X[n] * Y[n]);
	      }
	      
	      int[][] chosen = new int[K][N];
	 
	      double[] best = CalcMaxFast(fun, N, X);
	      chosen[0] = CalcMaxPoint(fun, N, X);
	      
	      double res = -1;

	      for(int k=0; k<K-2; k++) {
	        Function[] F= new Function[N];

	        for(int n=0; n < N; n ++) {
	          F[n] = new Function(-Y[n], X[n] * Y[n] + best[n]);
	        }

	        best = CalcMaxFast(F, N, X);
	        chosen[k+1] = CalcMaxPoint(F, N, X);
	      }
	    
	      
	      int startpos = 0;
	   
	      for(int n=0; n<N;n++) {
	    	   double help = fun[n].evaluate(0) + best[n];
	    	   if(help > res) {
	    		   res = help;
	    		   startpos = n;
	    	   }
	      }
	      
	      Point[] point = new Point[K];
	      point[K-1] = new Point(X[startpos], Y[startpos]);
	      
	      for(int k=K-2; k>=0; k--) {
	    	  startpos = chosen[k][startpos];
		      point[k] = new Point(X[startpos], Y[startpos]);
	      }

	      return point;
	    }

	   
	
/**
 * @param front : The front we want to select k points of
 * @param K  : The number of points to be selected from
 * @param refX : the x-coordinate of the reference point
 * @param refY : the y-coordinate of the reference point
 * @return A set of at most k points with the biggest hypervolume
 */
	public static Point[] calcPoints(Point[] front,  int K, Point reference ) {
			  int N = front.length;
			  
			  double refX = reference.x;
			  double refY = reference.y;
			  
			  for(int n=0; n<front.length; n++) {
				  front[n].x = refX - front[n].x;
				  front[n].y = refY - front[n].y;
			  }

			  // sort by x coordinate
		      Arrays.sort(front, new Comparator<Point>(){
		          @Override
		          public int compare(Point o1, Point o2) {
		              if (o1.x < o2.x) {
		              	return -1;
		              } else  if (o1.x == o2.x){
		            	  if(o1.y < o2.y) {
		            		  return -1;
		            	  } else if(o1.y == o2.y) {
		            		  return 0;
		            	  } else {
		            		  return 1;
		            	  }
		              } else {
		            	  return 1;
		              }
		          }
		      });
		      
		      
		      Stack<Integer> st = new Stack<Integer>();
		      
		      st.push(0);
		                  
		      for(int n=1; n<N; n++){
		    	  
		    	  while(st.size() > 0 && front[st.peek()].y <= front[n].y) {
		    		  st.pop();
		    	  } 
		    	  st.push(n);
		      }
		      
		      double X[] = new double[st.size()];
		      double Y[] = new double[st.size()];
		      
		      Function[] f = new Function[st.size()];
		      
		      for (int i=0; i < st.size(); i++) {
		        	int n = st.get(i);
			        double x= front[n].x;
			        double y= front[n].y;
			        X[i] = x;
			        Y[i] = y;
			        f[i] = new Function(-y, x*y);
		      }
		      		      	      
		      int size = st.size();
		      
		      Point[] res2;
		      if (Math.min(K, size) > 1 ) {
		    	  res2 = FasterPoint(size, Math.min(K, size), X, Y);	
		      } else { // special case K = 1 
		    	  Point p = new Point(0,0);
		    	  double best = 0;
		    	  for(int n=0; n< size; n++) {
		    		  if (X[n]*Y[n] > best) {
		    			  best = X[n]*Y[n];
		    			  p = new Point(X[n], Y[n]);
		    		  } 
		    	  }
		    	  
		    	  res2 = new Point[] {p};
		      }
		      // reconvert points
		      for(int n=0; n<res2.length; n++) {
		    	  res2[n].x = refX - res2[n].x;
		    	  res2[n].y = refY - res2[n].y;
		      }
		      
		      // reconvert points
		      for(int n=0; n<front.length; n++) {
				  front[n].x = refX - front[n].x;
				  front[n].y = refY - front[n].y;
			  }
		      
		      // reverse order
		      for(int n=front.length-1;n>=0; n--) {
		    	  int m = front.length-1-n;
		    	  if (n < m ){
		    		  double x1 = front[n].x;
		    		  double y1 = front[n].y;
		    		  front[n].x = front[m].x;
		    		  front[n].y = front[m].y;
		    		  front[m].x = x1;
		    		  front[m].y = y1;
		    		 }
		    	  }
		      
		      return res2;
		    }

	/**
	 * @param front : The front we want to select k points of
	 * @param K  : The number of points to be selected from
	 * @param refX : the x-coordinate of the reference point
	 * @param refY : the y-coordinate of the reference point
	 * @return the biggest value for HYPSSP
	 */
	  public static double calcValue(Point[] front,  int K, Point reference) {
		Point[] p = HYPSSP.calcPoints(front, K, reference);

		double refX = reference.x;
		double refY = reference.y;
		
		double ret =0;
		
		for (int n=p.length-1; n>=0; n--) {
			ret += (refX-p[n].x) * (refY-p[n].y);
			refX = p[n].x;
		}
		
		return ret;
	  }
	   
}
